import re
from typing import Union, List

from bs4 import BeautifulSoup, Tag, NavigableString

from hujiscrape.huji_objects import HujiObject, Lesson, Course, Exam

Bs4Obj = Union[BeautifulSoup, Tag, NavigableString]


class HtmlToObject:
    """
    Object that receives specific html and converts it to a huji object
    """

    def convert(self, html_str: str) -> HujiObject:
        raise NotImplementedError()


class HtmlToCourse(HtmlToObject):

    def _sort_lessons(self, lessons: List[Lesson]) -> List[Lesson]:
        """
        Returns a sorted copy of a list of lessons by semester, day, time, and other fields, to preserve the same
        order between different runs of the program.
        :param lessons: list of lessons
        :return: sorted list of lessons
        """
        return sorted(lessons, key=lambda lesson: (lesson.semester, lesson.day, lesson.start_time, lesson.end_time,
                                                   lesson.passing_type, lesson.location))

    def convert(self, html_str: str) -> Course:
        html = BeautifulSoup(html_str, 'html.parser')

        # Extract faculty and department
        faculty_data = html.find('div', class_='data-school').text.strip()
        faculty_parts = faculty_data.split(':')
        faculty = faculty_parts[0].strip()
        department = faculty_parts[1].strip() if len(faculty_parts) > 1 else ""

        # Extract course details from the title section
        course_id = html.find('div', class_='title').text.strip().split()[-1]
        hebrew_course_name = html.find('div', class_='subtitle').text.strip()
        english_course_name = html.find('div', class_='subtitle-eng').text.strip()

        # Extract additional data
        additional_data = html.find('div', class_='additional-data')
        semester = additional_data.find('div', class_='additional-data-semester').text.strip()
        # semester = Semester.from_hebrew(semester_text)

        weekly_hours_div = additional_data.find('div', class_='.additional-data-points')
        weekly_hours = int(re.search(r'\d+', weekly_hours_div.text.strip()).group()) if weekly_hours_div else 0

        credits_text = additional_data.find('div', class_='additional-data-student-points').text.strip()
        credit_points = int(re.search(r'\d+', credits_text).group())

        # Get test information
        test_div = additional_data.find('div', class_='additional-data-test')
        test_info = test_div.text.strip() if test_div else ""
        exam_type = test_info
        exam_length = 0
        if re.search(r'(\d+\.\d+)', test_info):
            exam_length = float(re.search(r'(\d+\.\d+)', test_info).group())

        # Get language
        language_div = additional_data.find('div', class_='additional-data-language')
        language = language_div.text.strip() if language_div else ""

        is_running = True  # Assuming course is running if it's in the system

        # Extract notes
        notes_div = html.find('div', id='comments-course-NumCourse')
        hebrew_notes = notes_div.text.strip() if notes_div else ""
        english_notes = ""  # Not found in the new format

        # Extract syllabus and moodle links
        syllabus_url = ""
        moodle_url = ""

        cyllabus_tag = html.select_one(".cyllabus-cource")
        if cyllabus_tag and "href" in cyllabus_tag.attrs:
            href = cyllabus_tag["href"]
            # Extract the URL from JavaScript function
            if "javascript:OpenUrl(" in href:
                url_parts = href.split("'")
                if len(url_parts) >= 2:
                    syllabus_url = f"https://shnaton.huji.ac.il{url_parts[1]}"

        moodle_tag = html.select_one(".moodle-cource")
        if moodle_tag and "href" in moodle_tag.attrs:
            moodle_url = moodle_tag["href"]

        # Extract schedule information
        schedule = []
        rows = html.find_all('div', class_='row')

        # Skip the title row
        for row_num, row in enumerate(rows[1:], 0):
            lecturer_name_div = row.find('div', class_='lecturer-name')
            lecturers = []
            if lecturer_name_div and lecturer_name_div.text.strip():
                lecturers = [line for line in lecturer_name_div.stripped_strings]

            # Get groups, semester, days, hours, lesson types
            groups = [group.text for group in row.find('div', class_='groups').find_all('div')] if row.find('div',
                                                                                                            class_='groups') else []
            semesters = [sem.text for sem in row.find('div', class_='semester').find_all('div')] if row.find('div',
                                                                                                             class_='semester') else []
            days = [day.text.strip() for day in row.find('div', class_='days').find_all('div', class_='day')] if row.find('div',
                                                                                                                  class_='days') else []
            hours = [hour.text for hour in row.find('div', class_='hour').find_all('div') if
                     hour.text.strip()] if row.find('div', class_='hour') else []
            lesson_types = [lesson.text for lesson in row.find('div', class_='lesson').find_all('div')] if row.find(
                'div', class_='lesson') else []
            places = [place.text.strip() for place in
                      row.find('div', class_='places').find_all('div', class_='place-item')] if row.find('div',
                                                                                                         class_='places') else []
            notes = [note.text.strip() for note in row.find('div', class_='note').find_all('div') if
                     note.text.strip()] if row.find('div', class_='note') else []

            # Create lesson objects
            max_items = max(len(semesters), len(days), len(hours), len(places))

            for i in range(max_items):
                group = groups[0] if groups else ""
                lesson_type = lesson_types[0] if lesson_types else ""
                semester_val = semesters[i] if i < len(semesters) else ""
                day_val = days[i] if i < len(days) else ""
                time_val = hours[i] if i < len(hours) else ""
                location_val = places[i] if i < len(places) else ""
                passing_type_val = notes[i] if i < len(notes) else ""

                lesson = Lesson(
                    location=location_val,
                    passing_type=passing_type_val,
                    time=time_val,
                    day=day_val,
                    semester=semester_val,
                    group=group,
                    type=lesson_type,
                    lecturers=lecturers,
                    row=row_num
                )
                schedule.append(lesson)

        # Sort the schedule
        schedule = self._sort_lessons(schedule)

        return Course(
            faculty=faculty,
            department=department,
            course_id=course_id,
            english_name=english_course_name,
            hebrew_name=hebrew_course_name,
            credits=credit_points,
            weekly_hours=weekly_hours,
            semester=semester,
            language=language,
            exam_length=exam_length,
            exam_type=exam_type,
            schedule=schedule,
            exams=None,  # We'll need to handle exams separately
            hebrew_notes=hebrew_notes,
            english_notes=english_notes,
            is_running=is_running,
            syllabus_url=syllabus_url,
            moodle_url=moodle_url
        )


class HtmlToExams(HtmlToObject):
    def convert(self, html_str: str) -> List[Exam]:
        html = BeautifulSoup(html_str, 'html.parser')
        exam_table = html.find('table')
        exams = []
        for tr in exam_table.find_all('tr')[1:]:
            semester, moed, exam_date, exam_hour, location, exam_notes = [td.text for td in tr.find_all('td')]
            exams.append(
                Exam(exam_date, exam_hour, exam_notes, location, moed, semester)
            )
        return exams



from dataclasses import dataclass
from functools import cached_property
from typing import List, Optional, Tuple


class HujiObject:
    pass


@dataclass(frozen=True)
class Lesson(HujiObject):
    location: str
    passing_type: str  # Meaning in campus / videotaped etc...
    time: str
    day: str
    semester: str
    group: str
    type: str  # Tirgul, Lecture...
    lecturers: List[str]

    # The row where the lesson appeared in the shnaton.
    # Helpful to know which lessons are considered the same (basically group + semester)
    row: int

    @cached_property
    def _split_time(self) -> Tuple[str, str]:
        """
        :return: self.time as a tuple of (start_time, end_time). If self.time == '', returns ('', '').
        """
        try:
            # The format of the time field is <end_time>-<start_time>
            end_time, start_time = self.time.split('-')
        except ValueError:
            return '', ''

        return start_time, end_time

    @property
    def start_time(self) -> str:
        return self._split_time[0]

    @property
    def end_time(self) -> str:
        return self._split_time[1]


@dataclass(frozen=True)
class Exam(HujiObject):
    date: str
    hour: str
    notes: str
    location: str
    moed: str
    semester: str


@dataclass(frozen=False)
class Course(HujiObject):
    course_id: str
    hebrew_name: str
    english_name: str
    department: str
    faculty: str
    semester: str
    weekly_hours: int
    credits: int
    language: str
    exam_length: float
    exam_type: str  # Project, written test, ...
    schedule: List[Lesson]
    exams: Optional[List[Exam]]
    hebrew_notes: str
    english_notes: str
    is_running: bool
    syllabus_url: str
    moodle_url: str



import asyncio
import random

import aiohttp

from hujiscrape.fetch_tasks import FetchTask

DEFAULT_MAX_CONCURRENCY = 100
DEFAULT_TCP_SOCKET_LIMIT = 20


class Fetcher:
    def __init__(self, session: aiohttp.ClientSession = None, retries: int = 3, timeout: aiohttp.ClientTimeout = None,
                 max_concurrency: int | None = DEFAULT_MAX_CONCURRENCY,
                 tcp_socket_limit: int = DEFAULT_TCP_SOCKET_LIMIT, force_close_tcp: bool = False):
        """
        Parameters are tuned to work with Shnaton scraping.
        :param force_close_tcp: There is a bug in aiohttp that causes the following error if force_close isn't true:
                                error type: <class 'aiohttp.client_exceptions.ClientOSError'>, error msg: [Errno None]
                                Can not write request body for URL
        """
        self._retries = retries
        self._session: aiohttp.ClientSession = session or aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit=tcp_socket_limit, force_close=force_close_tcp)
        )
        self._timeout = timeout or aiohttp.ClientTimeout(total=60, sock_connect=10, sock_read=10)
        self._semaphore = asyncio.BoundedSemaphore(max_concurrency) if max_concurrency else None

    async def __aenter__(self):
        await self._session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._session.__aexit__(exc_type, exc_val, exc_tb)

    async def fetch(self, task: FetchTask) -> str:
        if not self._semaphore:
            response = await self._fetch_with_retries(task)
        else:
            async with self._semaphore:
                response = await self._fetch_with_retries(task)

        # No need to bound the io on local response.
        return await response.text()

    async def _fetch_with_retries(self, task: FetchTask) -> aiohttp.ClientResponse:
        for attempt in range(1, self._retries + 1):
            try:
                response = await self._session.request(
                    method=task.method,
                    url=task.url,
                    data=task.data,
                    params=task.query_params,
                    headers=task.headers,
                    verify_ssl=False,
                    allow_redirects=False,
                    timeout=self._timeout,
                )
                response.raise_for_status()
                return response
            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                print(f"Request failed ({attempt}/{self._retries}): error type: {type(e)}, error msg: {e}")
                if attempt == self._retries:
                    raise e
                await asyncio.sleep(2 ** attempt + random.uniform(0, 1))  # Exponential backoff



import enum
from typing import Optional


class Prisa(enum.IntEnum):
    Minimal = 0
    Medium = 1
    Maximal = 2


# Type of degree
class Toar(enum.IntEnum):
    Any = 0
    Boger = 1


# Year in the toar
class ToarYear(enum.IntEnum):
    Any = 0
    First = 1
    Second = 2
    Third = 3
    Fourth = 4


# Semesters
class Semester(enum.IntEnum):
    """
    Numbers correlate to values used to filter in the Shnaton
    """
    # TODO: might need to change this to just a normal enum
    A = 1
    B = 2
    AB = 3
    Summer = 5
    Yearly = 9

    def __str__(self):
        return self.name

    @classmethod
    def from_hebrew(cls, text: str) -> Optional['Semester']:
        """
        A method that returns an enum value for a given text.
        :return: an enum
        :raise: ValueError if the text doesn't map to any enum.
        """
        semester = {
            "סמסטר א": Semester.A,
            "סמסטר א'": Semester.A,
            "סמסטר ב": Semester.B,
            "סמסטר ב'": Semester.B,
            "סמסטר א' או / ו ב'": Semester.AB,
            "קיץ": Semester.Summer,
            "סמסטר קיץ": Semester.Summer,
            "שנתי": Semester.Yearly,
        }.get(text.strip())

        if semester is None:
            raise ValueError(f"'{text}' is not a valid Semester")

        return semester


if __name__ == '__main__':
    a = Semester("סמסטר א'")
    print(a)



from hujiscrape.scrapers import (
    SingleCourseScraper
)

from hujiscrape.magics import (
    Toar,
    ToarYear,
    Semester
)

from hujiscrape.huji_objects import (
    HujiObject,
    Lesson,
    Course,
    Exam,
)

from hujiscrape.fetchers import (
    Fetcher
)

__all__ = [
    'Toar',
    'ToarYear',
    'Semester',
    'HujiObject',
    'Lesson',
    'Course',
    'Exam',
    'Fetcher',
    'SingleCourseScraper',
    'html_to_object',
    'scrapers',
    'fetchers',
    'fetch_tasks',
    'huji_objects',
    'magics',
]



import random


class FetchTask:
    def __init__(self, url: str, method: str, data: dict = None, query_params: dict = None,
                 headers: dict = None) -> None:
        self.url = url
        self.method = method
        self.data = data or {}
        self.query_params = query_params or {}
        self.headers = headers or {}


class ShnatonFetchTask(FetchTask):
    SHNATON_URL = 'https://shnaton.huji.ac.il/index.php'

    def __init__(self, data: dict = None, query_params: dict = None) -> None:
        super().__init__(url=self.SHNATON_URL, method='POST', data=data, query_params=query_params,
                         headers=self._get_default_headers())

    def _get_default_headers(self) -> dict:
        user_agents = [
            'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36',
            'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'
        ]
        return {
            'user-agent': random.choice(user_agents),
        }


class ExamFetchTask(ShnatonFetchTask):
    def __init__(self, course_id: int | str, year: int):
        super().__init__(
            data={
                'peula': 'CourseD',
                'year': year,
                'detail': 'examDates',
                'course': str(course_id)
            }
        )
        self.course_id = str(course_id)
        self.year = year


class CourseFetchTask(ShnatonFetchTask):
    def __init__(self, course_id: int | str, year: int):
        super().__init__(
            data={
                'peula': 'Simple',
                'maslul': 0,
                'shana': 0,
                'year': year,
                'course': course_id
            }
        )
        self.course_id = str(course_id)
        self.year = year



import asyncio
import logging
from functools import partial
from typing import List

from tqdm.asyncio import tqdm

from hujiscrape.fetch_tasks import CourseFetchTask, ExamFetchTask
from hujiscrape.fetchers import Fetcher
from hujiscrape.html_to_object import HtmlToCourse, HtmlToExams
from hujiscrape.huji_objects import Course


class ShnatonScraper:
    def __init__(self, fetcher: Fetcher | None = None) -> None:
        self._fetcher = fetcher or Fetcher()

    async def scrape(self, **kwargs):
        raise NotImplementedError()


class SingleCourseScraper(ShnatonScraper):
    MISSING_COURSE_TEXT = "לא נמצא קורס"

    def __init__(self, fetcher: Fetcher | None = None) -> None:
        """
        :param allow_missing_courses: should missing course requests result in an exception that stops the
                                      scraping process
        """
        super().__init__(fetcher)
        self._course_parser = HtmlToCourse()
        self._exam_parser = HtmlToExams()

    async def scrape(
            self,
            course_ids: list[int | str],
            year: int,
            include_exams: bool = True,
            show_progress: bool = False,
            fail_after_n_missing_courses: int = 0
    ) -> List[Course]:
        """
        Scrape multiple courses from shnaton
        :param course_ids: list of course ids to scrape
        :param year: year to scrape
        :param include_exams: should the exams be scraped as well
        :param show_progress: should a progress bar be shown
        :param fail_after_n_missing_courses: if more than n courses are missing, raise an exception.
                                             if 0, do not raise an exception
        :return: list of courses
        """
        courses = []
        failed_courses = 0

        async with self._fetcher:
            course_scrape_tasks = [
                self._scrape_single_course(CourseFetchTask(course_id, year))
                for course_id in course_ids
            ]

            await_for_courses = asyncio.as_completed if not show_progress else partial(tqdm.as_completed,
                                                                                       desc="Scraping courses")
            for course_coro in await_for_courses(course_scrape_tasks):
                course = await course_coro
                if course is None:
                    failed_courses += 1
                    if fail_after_n_missing_courses and failed_courses >= fail_after_n_missing_courses:
                        raise ValueError(f"Failed to fetch {failed_courses} courses")
                    continue

                courses.append(course)

            if include_exams:
                exam_scrape_tasks = [
                    self._attach_exams_to_course(course, year)
                    for course in courses
                ]

                # Separate await so that there is a progress bar for each
                await_for_exams = asyncio.as_completed if not show_progress else partial(tqdm.as_completed,
                                                                                         desc="Scraping exams")
                for exams_coro in await_for_exams(exam_scrape_tasks):
                    await exams_coro

            return courses

    async def _scrape_single_course(self, course_fetch_task: CourseFetchTask) -> Course | None:
        # Request course from shnaton
        course_html = await self._fetcher.fetch(course_fetch_task)

        if self.MISSING_COURSE_TEXT in course_html:
            logging.info(f"\nCourse {course_fetch_task.course_id} not found in {course_fetch_task.year}")

            return None

        # Parse the html to a course
        try:
            course = self._course_parser.convert(course_html)
        except Exception:
            print(f'Failed to convert course {course_fetch_task.course_id}')
            raise

        # If the course has an exam task, add it as well.
        # if course_fetch_task.exam_fetch_task is not None:
        #     exams_html = await self._fetcher.fetch(course_fetch_task.exam_fetch_task)
        #     exams = self._exam_parser.convert(exams_html)
        #     course.exams = exams

        return course

    async def _attach_exams_to_course(self, course: Course, year: int) -> None:
        exams_html = await self._fetcher.fetch(ExamFetchTask(course.course_id, year))
        course.exams = self._exam_parser.convert(exams_html)
